@@ -1,6 +1,7 @@
 /* cache flushing for the PPC
 
-  Copyright (C) 1998,2001,2003,2007 Free Software Foundation, Inc.
+  Authors: Anton Ertl, Bernd Paysan
+  Copyright (C) 1998,2001,2003,2007,2017,2019 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -18,6 +19,7 @@
   along with this program; if not, see http://www.gnu.org/licenses/.
 */
 
+#include "config.h"
 #include <stddef.h>
 #include <sys/types.h>
 
