@@ -1,7 +1,7 @@
 /* Android main() for Gforth on Android
 
   Authors: Bernd Paysan, Anton Ertl
-  Copyright (C) 2012,2013,2014,2015,2016,2017,2018,2019 Free Software Foundation, Inc.
+  Copyright (C) 2012,2013,2014,2015,2016,2017,2018,2019,2020,2021 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -29,6 +29,57 @@
 #include <fcntl.h>
 #include <stdarg.h>
 #include <pthread.h>
+#include <errno.h>
+
+#if defined(__LP64__)
+struct __real_sbuf {
+  unsigned char* _base;
+  size_t _size;
+};
+#else
+struct __real_sbuf {
+        unsigned char *_base;
+        int     _size;
+};
+#endif
+typedef struct __real_sFILE {
+        unsigned char *_p;      /* current position in (some) buffer */
+        int     _r;             /* read space left for getc() */
+        int     _w;             /* write space left for putc() */
+#if defined(__LP64__)
+        int     _flags;         /* flags, below; this FILE is free if 0 */
+        int     _file;          /* fileno, if Unix descriptor, else -1 */
+#else
+        short   _flags;         /* flags, below; this FILE is free if 0 */
+        short   _file;          /* fileno, if Unix descriptor, else -1 */
+#endif
+        struct  __real_sbuf _bf;     /* the buffer (at least 1 byte, if !NULL) */
+        int     _lbfsize;       /* 0 or -_bf._size, for inline putc */
+
+        /* operations */
+        void    *_cookie;       /* cookie passed to io functions */
+        int     (*_close)(void *);
+        int     (*_read)(void *, char *, int);
+        fpos_t  (*_seek)(void *, fpos_t, int);
+        int     (*_write)(void *, const char *, int);
+
+        /* extension data, to avoid further ABI breakage */
+        struct  __real_sbuf _ext;
+        /* data for long sequences of ungetc() */
+        unsigned char *_up;     /* saved _p when _p is doing ungetc data */
+        int     _ur;            /* saved _r when _r is counting ungetc data */
+
+        /* tricks to meet minimum requirements even when malloc() fails */
+        unsigned char _ubuf[3]; /* guarantee an ungetc() buffer */
+        unsigned char _nbuf[1]; /* guarantee a getc() buffer */
+
+        /* separate buffer for fgetln() when line crosses buffer boundary */
+        struct  __real_sbuf _lb;     /* buffer for fgetln() */
+
+        /* Unix stdio files get aligned to block boundaries on fseek() */
+        int     _blksize;       /* stat.st_blksize (may be != _bf._size) */
+        fpos_t  _offset;        /* current lseek offset */
+} real_FILE;
 
 #include "forth.h"
 
@@ -58,6 +109,7 @@ typedef struct {
   char* libdir;
   char* locale;
   char* startfile;
+  char** filedirs;
 } jniargs;
 
 jniargs startargs;
@@ -100,8 +152,7 @@ int checksha256sum(const char * sha256, const char* sha256file)
     LOGI("cksha256: size %d wrong\n", checkread);
     return 0;
   }
-  if(memcmp(sha256buffer, sha256, 64)) return 0;
-  return 1;
+  return !memcmp(sha256buffer, sha256, 64);
 }
 
 void post(char * doprog)
@@ -150,13 +201,20 @@ char *get_gforth_gz()
 
 int unpackFiles(const char* zipfile, const char* sumfile, const char* shasum)
 {
-  int csfd, writeout;
+  int csfd, writeout=0;
   post("showprog");
   zexpand(zipfile);
-  csfd=creat(sumfile, O_WRONLY);
-  writeout=write(csfd, shasum, 64);
-  fprintf(stderr, "sha256sum: '%64s'\n", shasum);
-  close(csfd);
+  csfd=open(sumfile, O_WRONLY | O_CREAT, 0644);
+  if(csfd<0) {
+    LOGE("can't create \"%s\": %s\n", sumfile, strerror(errno));
+    unlink(sumfile);
+    csfd=open(sumfile, O_WRONLY | O_CREAT, 0644);
+  }
+  if(csfd>=0) {
+    writeout=write(csfd, shasum, 64);
+    fprintf(stderr, "sha256sum: '%64s'\n", shasum);
+    close(csfd);
+  }
   if(writeout!=64) {
     post("errprog");
     return 0;
@@ -194,53 +252,66 @@ void addfileargs(char* filename)
   if(argfile==NULL) return; // no file, nothing to do
 
   while((line=fgetln(argfile, &n))) {
-    if(n > 0 && line[n-1]=='\n') {
-      n--;
-      addarg(line, n);
-    }
+    // remove CR/LF line endings
+    if(n > 0 && line[n-1]=='\n') n--;
+    if(n > 0 && line[n-1]=='\r') n--;
+    // only add non-empty lines
+    if(n > 0) addarg(line, n);
   }
 }
 
-static const char *paths[] = { "--",
-			       "--path=.:/mnt/sdcard/gforth/current:/mnt/sdcard/gforth/" ARCH "/gforth/current:/mnt/sdcard/gforth/site-forth:/mnt/sdcard/gforth/" ARCH "/gforth/site-forth",
-			       "--path=.:/data/data/gnu.gforth/files/gforth/current:/data/data/gnu.gforth/files/gforth/" ARCH "/gforth/current:/data/data/gnu.gforth/files/gforth/site-forth:/data/data/gnu.gforth/files/gforth/" ARCH "/gforth/site-forth" };
-static const char *folder[] = { "/sdcard", "/mnt/sdcard", "/data/data/gnu.gforth/files" };
+char *paths;
 char *rootdir;
 char *homedir;
 
-int checkFiles(char ** patharg)
+int checkFiles(char ** patharg, char * filedirs[])
 {
-  int i;
+  int i, shacheck=0;
   FILE * test;
   char * logfile;
 
-  for(i=0; i<=2; i++) {
-    *patharg=paths[i];
-    if(!chdir(folder[i])) {
-      mkdir("gforth", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
-      // may fail, because it exists
-      if((test=fopen("gforth/test-stamp", "w+"))) {
-	fclose(test);
-	unlink("gforth/test-stamp");
-	LOGI("chdir(%s)\n", folder[i]);
-	break;
+  for(i=0; filedirs[i]!=NULL; i++) {
+    LOGI("folder[%i] = \"%s\"\n", i, filedirs[i]);
+    if(!chdir(filedirs[i])) {
+      LOGI("chdir(%s)+shacheck\n", filedirs[i]);
+      // check if the files already have been successfully unpacked
+      if((shacheck =
+	  checksha256sum(sha256sum, "gforth/current/sha256sum") &&
+	  checksha256sum(sha256arch, "gforth/" ARCH "/gforth/current/sha256sum"))) break;
+    }
+  }
+  // if this failed, check if you can create a directory
+  // and write into that directory
+  if(!shacheck) {
+    for(i=0; filedirs[i]!=NULL; i++) {
+      LOGI("try create folder[%i] = \"%s\"\n", i, filedirs[i]);
+      if(!chdir(filedirs[i])) {
+	mkdir("gforth", S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
+	if((test=fopen("gforth/test-stamp", "w+"))) {
+	  fclose(test);
+	  unlink("gforth/test-stamp");
+	  LOGI("chdir(%s)+test-stamp succeeded\n", filedirs[i]);
+	  break;
+	}
       }
     }
   }
+  
+  rootdir=filedirs[i];
 
-  rootdir=folder[i];
-  LOGI("Extra arg: %s\n", *patharg);
+  asprintf(&paths,
+	   "--path=.:%s/gforth/current:%s/gforth/" ARCH "/gforth/current:%s/gforth/site-forth:%s/gforth/" ARCH "/gforth/site-forth",
+	   rootdir, rootdir, rootdir, rootdir);
+  LOGI("Extra arg: %s\n", paths);
+  *patharg=paths;
 
-  if(i != 0) {
-    asprintf(&logfile, "%s/gforthout.log", rootdir);
-    freopen(logfile, "w+", stdout);
-    asprintf(&logfile, "%s/gfortherr.log", rootdir);
-    freopen(logfile, "w+", stderr);
-    free(logfile);
-  }
+  asprintf(&logfile, "%s/gforthout.log", rootdir);
+  freopen(logfile, "w+", stdout);
+  asprintf(&logfile, "%s/gfortherr.log", rootdir);
+  freopen(logfile, "w+", stderr);
+  free(logfile);
 
-  return checksha256sum(sha256sum, "gforth/current/sha256sum") &&
-    checksha256sum(sha256arch, "gforth/" ARCH "/gforth/current/sha256sum");
+  return shacheck;
 }
 
 void startForth(jniargs * startargs)
@@ -255,13 +326,13 @@ void startForth(jniargs * startargs)
   JavaVMAttachArgs vmAA = { JNI_VERSION_1_6, "NativeThread", 0 };
 
   pipe(epipe);
-  stdin->_file=epipe[0];
+  ((real_FILE*)(stdin))->_file=epipe[0];
 
   (*vm)->AttachCurrentThread(vm, &env, &vmAA);
   
   startargs->env = env;
 
-  if(!checkFiles(&patharg)) {
+  if(!checkFiles(&patharg, startargs->filedirs)) {
     char *dir = startargs->libdir;
     const char *zip = "libgforth-" ARCH "gz.so";
     int  dirlen=strlen(dir)+strlen(zip)+2;
@@ -278,14 +349,12 @@ void startForth(jniargs * startargs)
     free(gforth_gz);
   }
 
-  if(rootdir) {
+  if(strcmp(rootdir, "/sdcard")) {
     asprintf(&homedir, "%s/gforth/home", rootdir);
     setenv("HOME", homedir, 1);
     free(homedir);
-    if((rootdir != folder[0])) {
-      setenv("GFORTHDESTDIR", folder[0], 1);
-      setenv("GFORTHINSDIR", rootdir, 1);
-    }
+    setenv("GFORTHDESTDIR", "/sdcard", 1);
+    setenv("GFORTHINSDIR", rootdir, 1);
   } else {
     setenv("HOME", "/sdcard/gforth/home", 1);
   }
@@ -313,7 +382,6 @@ void startForth(jniargs * startargs)
     if(needhide) {
       post("hideprog");
     }
-    gforth_setwinch();
     gforth_bootmessage();
     LOGI("starting gforth_quit\n");
     retvalue = gforth_quit();
@@ -337,13 +405,20 @@ pthread_attr_t * pthread_detach_attr(void)
   return &attr;
 }
 
-void JNI_startForth(JNIEnv * env, jobject obj, jstring libdir, jstring locale, jstring startfile)
+void JNI_startForth(JNIEnv * env, jobject obj, jstring libdir, jstring locale, jstring startfile, jobjectArray filedirs)
 {
+  int i, filedirsnum = (*env)->GetArrayLength(env, filedirs);
   startargs.obj = (*env)->NewGlobalRef(env, obj);
   startargs.win = 0; // is a native window
   startargs.libdir = getjstring(env, libdir);
   startargs.locale = getjstring(env, locale);
   startargs.startfile = getjstring(env, startfile);
+  startargs.filedirs = malloc((filedirsnum+1)*sizeof(char*));
+  
+  for(i=0; i<filedirsnum; i++) {
+    startargs.filedirs[i] = getjstring(env, (jstring)(*env)->GetObjectArrayElement(env, filedirs, i));
+  }
+  startargs.filedirs[filedirsnum] = NULL;
 
   pthread_create(&(startargs.id), pthread_detach_attr(), startForth, (void*)&startargs);
 }
@@ -381,7 +456,7 @@ static JNINativeMethod GforthMethods[] = {
   {"onEventNative", "(ILjava/lang/Object;)V", (void*) JNI_onEventNative},
   {"onEventNative", "(II)V",                  (void*) JNI_onEventNativeInt},
   {"callForth",     "(J)V",                   (void*) JNI_callForth},
-  {"startForth",    "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V",  (void*) JNI_startForth},
+  {"startForth",    "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/String;)V",  (void*) JNI_startForth},
 };
 
 #define alen(array)  sizeof(array)/sizeof(array[0])
@@ -395,8 +470,8 @@ JNIEXPORT jint JNI_OnLoad(JavaVM* vm, void* reserved)
   
   LOGI("Enter onload\n");
   
-  freopen("/sdcard/gforthout.log", "w+", stdout);
-  freopen("/sdcard/gfortherr.log", "w+", stderr);
+  // freopen("/sdcard/gforthout.log", "w+", stdout);
+  // freopen("/sdcard/gfortherr.log", "w+", stderr);
 
   pipe(startargs.ke_fd);
 
