@@ -2,7 +2,7 @@
 
 
   Author: Bernd Paysan
-  Copyright (C) 2012,2019 Free Software Foundation, Inc.
+  Copyright (C) 2012,2019,2024 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -22,8 +22,22 @@
 
 #include "config.h"
 #include "forth.h"
+#include <stdlib.h>
+#include <signal.h>
 
 int main(int argc, char **argv, char **env)
 {
-  return gforth_main(argc, argv, env);
+  int retval=gforth_main(argc, argv, env);
+  switch(retval) {
+  case 0: return EXIT_SUCCESS;
+  case -9:  return 0x80|SIGSEGV;
+  case -28: return 0x80|SIGINT;
+  case -55: return 0x80|SIGFPE;
+#ifdef SIGPIPE
+  case -2049: return 0x80|SIGPIPE;
+#endif
+  case -0x11F ... -0x100: return 0x80|((-retval) & 0x1F);
+  case 1 ... 0xFF: return retval;
+  default: return EXIT_FAILURE;
+  }
 }
